# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import functools
import itertools
import operator
from collections.abc import Generator

import networkx as nx
import pulp as lp

from slambuc.alg import T_PART, T_BARRS
from slambuc.alg.service.common import *


def verify_limits(tree: nx.DiGraph, cpath: set[int], M: int | float, L: int | float) -> tuple[bool, bool]:
    """
    Verify that given limits *M* and *L* based on the given *tree* allow feasible solution.

    :param tree:    input tree
    :param cpath:   nodes of critical path
    :param M:       memory upper bound
    :param L:       latency upper bound
    :return:        boolean values of satisfied limits M and L
    """
    return (max(tree.nodes[v][MEMORY] for v in tree if v is not PLATFORM) <= M,
            sum(tree.nodes[v][RUNTIME] for v in cpath) <= L)


def ipostorder_dfs(tree: nx.DiGraph, source: int, inclusive: bool = True) -> Generator[tuple[int, int]]:
    """
    Return the existing predecessor and node tuple in a DFS traversal of the given *tree* in a post/reversed order.

    :param tree:        input tree
    :param source:      starting node
    :param inclusive:   also return the source node
    :return:            generator of tree nodes
    """
    stack = collections.deque([(source, iter(tree[source]))])
    while stack:
        v, ichildren = stack[-1]
        try:
            c = next(ichildren)
            stack.append((c, iter(tree[c])))
        except StopIteration:
            stack.pop()
            if stack:
                yield stack[-1][0], v
    if inclusive:
        yield next(tree.predecessors(source), None), source


def ipostorder_tabu_dfs(tree: nx.DiGraph, source: int, tabu: set = None,
                        inclusive: bool = True) -> Generator[tuple[int, int]]:
    """
    Return nodes of *tree* in a postorder DFS traversal excluding descendants of nodes in *tabu* set.

    :param tree:        input tree
    :param source:      starting node
    :param tabu:        tabu node set
    :param inclusive:   also return the source node
    :return:            generator of tree nodes
    """
    stack = collections.deque([(source, iter(tree[source]))])
    while stack:
        v, ichildren = stack[-1]
        try:
            c = next(ichildren)
            if not tabu or c not in tabu:
                stack.append((c, iter(tree[c])))
        except StopIteration:
            stack.pop()
            if stack:
                yield stack[-1][0], v
    if inclusive:
        yield next(tree.predecessors(source), None), source


def ipostorder_edges(tree: nx.DiGraph, source: int, data: bool = False) -> Generator[tuple[int, int]]:
    """
    Return the edges (head, tail) in a DFS traversal of the given *tree* in a post/reversed order with edge data.

    :param tree:        input tree
    :param source:      starting node
    :param data:        return edge data
    :return:            generator of edges
    """
    stack = collections.deque([(source, iter(tree[source]))])
    while stack:
        v, ichildren = stack[-1]
        try:
            c = next(ichildren)
            stack.append((c, iter(tree[c])))
        except StopIteration:
            stack.pop()
            if stack:
                p = stack[-1][0]
                if data:
                    yield p, v, tree[p][v]
                else:
                    yield p, v


def ileft_right_dfs(tree: nx.DiGraph, source: int) -> Generator[tuple[tuple[int, int], int, tuple[int, int]]]:
    """
    Return edges in left-right traversal along with the previously visited uncle and sibling edges.

    :param tree:    input tree
    :param source:  starting node
    :return:        generator of interested nodes
    """
    # ... -> [u] -j-> [v] -i-> [v_i]  => [(u, u_j-1), v_i-1, v_succ, v, v_i]
    stack = collections.deque([[(None, None), None, iter(tree[source]), source, 0]])
    while stack:
        u_prev, v_prev, ichildren, v, v_i = stack[-1]
        yield u_prev, v_prev, (v, v_i)
        try:
            c = next(ichildren)
            stack[-1][1], stack[-1][-1] = v_i, c
            stack.append([(v, v_i), None, iter(tree[c]), c, 0])
        except StopIteration:
            stack.pop()


def ileft_right_dfs_idx(tree: nx.DiGraph, source: int) -> Generator[tuple[int, int]]:
    """
    Return nodes of the given *tree* in left-right traversal along with the index of the considered child node.

    :param tree:    input tree
    :param source:  starting node
    :return:        generator of interested node indices
    """
    # [v] -i-> [v_i]  => [v_succ, v, v_i]
    stack = collections.deque([[iter(tree[source]), source, 0]])
    while stack:
        ichildren, v, i = stack[-1]
        yield v, i
        try:
            c = next(ichildren)
            stack[-1][-1] += 1
            stack.append([iter(tree[c]), c, 0])
        except StopIteration:
            stack.pop()


def ichain(tree: nx.DiGraph, start: int, leaf: int) -> Generator[int]:
    """
    Generator over the nodes of the chain from *start* node to *leaf* node.

    :param tree:    input tree
    :param start:   first node
    :param leaf:    last node
    :return:        generator of chain nodes
    """
    n = start
    while n != leaf:
        yield n
        for c in tree.successors(n):
            if leaf in tree.nodes[c][LABEL]:
                n = c
                break
    yield leaf


def ibacktrack_chain(tree: nx.DiGraph, start: int, leaf: int) -> Generator[int]:
    """
    Return the node of a chain in the *tree* in backward order from *leaf* to *start* node.

    :param tree:    input tree
    :param start:   first node
    :param leaf:    last node
    :return:        generator of chain nodes
    """
    if leaf not in tree:
        return
    last = leaf
    while last != start:
        yield last
        try:
            last = next(tree.predecessors(last))
        except StopIteration:
            break
    yield last


def isubchains(tree: nx.DiGraph, start: int, leaf: int = None) -> Generator[tuple[list[int], list[int]], set[int]]:
    """
    Generator over the subchains and its branches from *start* to all reachable leaf where the subchain is bisected
    at the last node from which the specific *leaf* is still reachable.

    :param tree:    input tree
    :param start:   first node
    :param leaf:    last node
    :return:        generator of chain node parts and branches
    """
    chain = [start]
    while (deg := len(tree.succ[chain[-1]])) == 1:
        chain.append(next(tree.successors(chain[-1])))
    if deg == 0:
        yield (chain, []), set()
    else:
        for c in (children := set(tree.successors(chain[-1]))):
            nbr = children - {c}
            for (part1, part2), branches in isubchains(tree, c, leaf):
                if leaf in tree.nodes[part1[0]][LABEL]:
                    yield (chain + part1, part2), nbr | branches
                elif leaf is not None:
                    yield (chain, part1 + part2), nbr | branches
                else:
                    yield (chain + part1, []), nbr | branches


def iflattened_tree(tree: nx.DiGraph, root: int) -> Generator[list[int]]:
    """
    Generate chain decomposition of the given *tree* started from node *root*.

    :param tree:    input tree
    :param root:    root node
    :return:        generator of decomposed chains
    """
    for (head, tail), brs in isubchains(tree, root, max(tree.nodes[root][LABEL])):
        chain, brs = [head + tail], sorted(brs)
        for subchains in itertools.product(*map(functools.partial(iflattened_tree, tree), brs)):
            yield list(itertools.chain(chain, *subchains))


def isubtree_bfs(tree: nx.DiGraph, root: int) -> Generator[int]:
    """
    Return nodes in BFS traversal of the given *tree* started from *root*.

    :param tree:    input tree
    :param root:    root node
    :return:        generator of tree nodes
    """
    children = collections.deque((root,))
    yield root
    while children:
        u = children.popleft()
        for v in tree.successors(u):
            children.append(v)
            yield v


def isubtrees(tree: nx.DiGraph, barrs: T_BARRS) -> Generator[tuple[int, list[int]]]:
    """
    Return the barrier nodes and subtrees of the given *tree* marked by the *barr* nodes.

    :param tree:    input tree
    :param barrs:   set of barrier nodes
    :return:        generator of barrier and regarding subtree nodes
    """
    for b in barrs:
        nodes = [b]
        children = collections.deque(nodes)
        while children:
            u = children.popleft()
            for v in tree.successors(u):
                if v not in barrs:
                    nodes.append(v)
                    children.append(v)
        nodes.sort()
        yield b, nodes


def itop_subtree_nodes(tree: nx.DiGraph, root: int, barrs: T_BARRS) -> Generator[int]:
    """
    Return the nodes of the top subtree with *root* of the given *tree* cut by the *barr* nodes.

    :param tree:    input tree
    :param root:    root node
    :param barrs:   set of barrier nodes
    :return:        generator of topmost block's nodes
    """
    nodes = [root]
    children = collections.deque(nodes)
    while children:
        u = children.popleft()
        for v in tree.successors(u):
            if v not in barrs:
                children.append(v)
                yield v


def induced_subtrees(tree: nx.DiGraph, root: int,
                     only_nodes: bool = False) -> Generator[tuple[tuple[int, int], list[int | tuple[int]]]]:
    """
    Recursively generate the ingress edge of subtrees and all reachable edges / nodes in the given subtree.

    :param tree:        input tree
    :param root:        root node
    :param only_nodes:  return only subtree nodes instead of edges
    :return:            generator of ingress and covered edges
    """
    subtrees = collections.defaultdict(list)
    for p, n in ipostorder_dfs(tree, root, inclusive=True):
        base = (n,) if only_nodes else ((n, s) for s in tree.successors(n))
        subtrees[n].extend(itertools.chain(base, *(subtrees[c] for c in tree.successors(n))))
        for c in tree.successors(n):
            del subtrees[c]
        yield (p, n), subtrees[n]


def ipowerset(data: list[int], start: int = 0) -> Generator[list[int]]:
    """
    Generate the powerset of the given *data* beginning to count the sets from size *start*.

    :param data:    list of data values
    :param start:   lower bound of set size
    :return:        generator of subsets
    """
    return itertools.chain.from_iterable(itertools.combinations(data, i) for i in range(start, len(data) + 1))


def iser_mul_factor(rate: list[int]) -> Generator[int]:
    """
    Generator over the **pessimistic** number of function instances inside a block assuming serialized execution model.

    :param rate:    list of rate values
    :return:        generator of accumulated function instance count
    """
    return itertools.accumulate((math.ceil(j / i) for i, j in itertools.pairwise(rate)), operator.mul, initial=1)


def ipar_mul_factor(rate: list[int], N: int = 1) -> Generator[int]:
    """
    Generator over the **pessimistic** number of function instances inside a block assuming parallelized execution model.

    :param rate:    list of rate values
    :param N:       CPU count
    :return:        generator of accumulated function instance count
    """
    return itertools.accumulate((math.ceil(j / (i * N)) for i, j in itertools.pairwise(rate)), operator.mul, initial=1)


def igen_mul_factor(rate: list[int], ncores: list[int]) -> Generator[int]:
    """
    Generator over the **pessimistic** number of function instances using separate relative CPU cores.

    :param rate:    list of rate values
    :param ncores:  list of CPU cores
    """
    return itertools.accumulate((math.ceil(j / (i * nc)) for (i, j), nc in zip(itertools.pairwise(rate), ncores)),
                                operator.mul, initial=1)


########################################################################################################################


def leaf_label_nodes(tree: nx.DiGraph) -> nx.DiGraph:
    """
    Label each node *n* with the set of leafs that can be reached from *n*.

    :param tree:    input tree
    :return:        labeled tree
    """
    for _, v in ipostorder_dfs(tree, PLATFORM, inclusive=False):
        tree.nodes[v][LABEL] = set().union(*(tree.nodes[s][LABEL] for s in tree.succ[v])) if len(tree.succ[v]) else {v}
    return tree


def ith_child(tree: nx.DiGraph, v: int, i: int) -> int:
    """
    Returns the *i*-th child of the node *v* started to count from 1.

    E.g.:
    >>> v_i = ith_child(tree, v, i) # [v] -i-> [v_i]

    :param tree:    input tree
    :param v:       node ID
    :param i:       number of given child
    :return:        i-th node
    """
    return next(itertools.islice(tree.successors(v), i - 1, None)) if i > 0 else 0


def child_idx(tree: nx.DiGraph, v: int) -> int:
    """
    Returns the index of *v* among its sibling nodes or return 0.

    E.g.:
    >>> j = child_idx(tree, v) # [u] -j-> [v]

    :param tree:    input tree
    :param v:       node ID
    :return:        index of node *v*
    """
    return next((i for i, v_i in enumerate(tree.succ[next(tree.predecessors(v))], 1) if v_i == v)) if v else 0


def top_subtree_block(tree: nx.DiGraph, barr: T_BARRS) -> nx.DiGraph:
    """
    Return the first/top subtree of the given *tree* separated by the given *barr* nodes.

    :param tree:    input tree
    :param barr:    set of barrier nodes
    :return:        top subtree
    """
    return tree.subgraph(next(isubtrees(tree, sorted(barr)))[1])


def path_blocks(partition: T_PART, path: list[int]) -> T_PART:
    """
    Calculate the blocks of separated critical path based on the original partitioning.

    :param partition:   given tree partitioning
    :param path:        path of specific nodes
    :return:            calculated path blocks
    """
    parts = []
    current_blk = None
    for v in path:
        for blk in partition:
            if v in blk:
                if blk == current_blk:
                    parts[-1].append(v)
                else:
                    parts.append([v])
                    current_blk = blk
    return parts


def recreate_subchain_blocks(tree: nx.DiGraph, barr: T_BARRS) -> T_PART:
    """
    Recreate chain blocks from barrier nodes of the given partitioning.

    :param tree:    input tree
    :param barr:    set of barrier nodes
    :return:        list of chain blocks
    """
    n = list(tree.nodes)
    p = []
    if not barr:
        return []
    for w in reversed(range(1, len(n))):
        if n[w] == 0:
            continue
        n[w], blk, v = 0, [w], w
        while v not in barr:
            v = next(tree.predecessors(v))
            if n[v]:
                n[v] = 0
                blk.append(v)
        blk.reverse()
        p.append(blk)
    return sorted(p)


def recreate_subtree_blocks(tree: nx.DiGraph, barr: T_BARRS) -> T_PART:
    """
    Return the partition blocks of the given *tree* cut by the *barr* nodes.

    :param tree:    input tree
    :param barr:    set of barrier nodes
    :return:        list of partition blocks
    """
    p = []
    for b in barr:
        blk = [b]
        c = collections.deque(blk)
        while c:
            u = c.popleft()
            for v in tree.successors(u):
                if v not in barr:
                    blk.append(v)
                    c.append(v)
        p.append(blk)
    return sorted(p)


def split_chain(barr: T_BARRS, n: int, full: bool = True) -> T_PART:
    """
    Recreate partition blocks from barrier nodes for an *n*-size chain := [0, n-1].

    :param barr:    set of barrier nodes
    :param n:       chain size
    :param full:    recreate all block nodes instead of just fist/last nodes
    :return:        created partitioning
    """
    return [list(range(b, w)) if full else [b, w - 1] if b < w - 1 else [b]
            for b, w in itertools.pairwise(barr + [n])]


def split_path(path: list[int], barr: T_BARRS) -> T_PART:
    """
    Recreate partition blocks of a chain from barrier nodes for an *n*-size chain := [0, n-1].

    :param path:    list of nodes
    :param barr:    set of barrier nodes
    :return:        created partitioning
    """
    return [path[b: w] for b, w in itertools.pairwise(barr + [len(path)])]


def x_eval(x: int | None | lp.LpVariable) -> bool:
    """
    Evaluate *x* from a decision variable matrix based on its solution value.

    :param x:   decision variable
    :return:    whether it is a solution or not
    """
    return bool(round(x.varValue) if isinstance(x, lp.LpVariable) else x)


def recalculate_ser_partitioning(tree: nx.DiGraph, partition: T_PART, root: int = 1, cp_end: int = None,
                                 delay: int = 1) -> tuple[int, int]:
    """
    Calculate the sum cost and sum latency on the critical path based on the given *partition* assuming serialized
    execution model.

    :param tree:        input tree
    :param partition:   given partitioning
    :param root:        root node
    :param cp_end:      end node of critical path
    :param delay:       platform invocation delay
    :return:            sum cost nad latency of the given partitioning
    """
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    c_blks = [blk for blk in partition if blk[0] in cpath]
    return (sum(ser_subtree_cost(tree, blk[0], blk) for blk in partition),
            sum(ser_subchain_latency(tree, blk[0], set(blk), cpath) for blk in c_blks) + (len(c_blks) - 1) * delay)


def recalculate_partitioning(tree: nx.DiGraph, partition: T_PART, root: int = 1, N: int = 1, cp_end: int = None,
                             delay: int = 1) -> tuple[int, int]:
    """
    Calculate sum cost and sum latency on the critical path based on the given *partition* assuming parallelized
    execution model.

    :param tree:        input tree
    :param partition:   given partitioning
    :param root:        root node
    :param N:           CPU count
    :param cp_end:      end node of critical path
    :param delay:       platform invocation delay
    :return:            sum cost nad latency of the given partitioning
    """
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    c_blks = [blk for blk in partition if blk[0] in cpath]
    return (sum(par_subtree_cost(tree, min(blk), blk, N) for blk in partition),
            sum(par_subchain_latency(tree, blk[0], set(blk), cpath, N) for blk in c_blks) + (len(c_blks) - 1) * delay
            if cpath else None)


########################################################################################################################


def block_memory(memory: list[int], b: int, w: int) -> int:
    """
    Calculate cumulative memory of block [b, w].

    :param memory:  list of memory values
    :param b:       barrier node
    :param w:       end node of block
    :return:        memory value
    """
    return sum(itertools.islice(memory, b, w + 1))


def block_cpu(rate: list[int], b: int, w: int) -> int:
    """
    Calculate CPU core need of block [b, w] with multiprocessing.

    :param rate:    list of rate values
    :param b:       barrier node
    :param w:       end node of block
    :return:        CPU count
    """
    r_max = itertools.chain((1,), enumerate(itertools.accumulate(reversed(rate[b: w + 1]), max)))
    return functools.reduce(lambda pre, max_i: max(pre, math.ceil(max_i[1] / rate[w - max_i[0]])), r_max)


def block_cost(runtime: list[int], rate: list[int], b: int, w: int, unit: int = 100) -> int:
    """
    Calculate running time of block [b, w] with multiprocessing.

    :param runtime: list of runtime values
    :param rate:    list of rate values
    :param b:       barrier node
    :param w:       end node of block
    :param unit:    rounding unit
    :return:        calculated cost
    """
    return rate[b] * (math.ceil(sum(runtime[b: w + 1]) / unit) * unit)


def block_latency(runtime: list[int], b: int, w: int, delay: int, start: int, end: int) -> int:
    """
    Calculate relevant latency for block [b, w] with multiprocessing.

    :param runtime: list of runtime values
    :param b:       barrier node
    :param w:       end node of block
    :param delay:   platform delay
    :param start:   fist node to consider
    :param end:     last node to consider
    :return:        calculated latency value
    """
    if end < b or w < start:
        # Do not consider latency if no intersection
        return 0
    blk_lat = sum(runtime[max(b, start): min(w, end) + 1])
    # Ignore delay if latency path starts within the subchain
    return delay + blk_lat if start < b else blk_lat


########################################################################################################################


def ser_block_memory(memory: list[int]) -> int:
    """
    Calculate cumulative memory of block [b, w] with serialization.

    :param memory:  list of memory values
    :return:        memory value
    """
    return sum(memory)


def ser_block_memory_opt(memory: list[int], rate: list[int], b: int, w: int) -> int:
    """
    Calculate memory of block [b, w] recursively based on the **optimistic** number of parallel function instances.

    :param memory:  list of memory values
    :param rate:    list of rate values
    :param b:       barrier node
    :param w:       end node of block
    :return:        calculated memory value
    """
    return functools.reduce(lambda pre, i: max(pre, max(math.ceil(rate[j] / rate[i]) * memory[j]
                                                        for j in range(i, w + 1))), reversed(range(b, w + 1)), 0)


def ser_block_memory_pes(memory: list[int], rate: list[int], b: int, w: int) -> int:
    """
    Calculate memory of block [b, w] recursively based on the **pessimistic** number of parallel function instances.

    :param memory:  list of memory values
    :param rate:    list of rate values
    :param b:       barrier node
    :param w:       end node of block
    :return:        calculated memory value
    """
    return functools.reduce(lambda pre, i: max(memory[i], math.ceil(rate[i + 1] / rate[i]) * pre),
                            reversed(range(b, w)), memory[w])


def ser_block_memory_pes2(memory: list[int], rate: list[int], b: int, w: int) -> int:
    """
    Calculate memory of block [b, w] directly based on the **pessimistic** number of parallel function instances.

    :param memory:  list of memory values
    :param rate:    list of rate values
    :param b:       barrier node
    :param w:       end node of block
    :return:        calculated memory value
    """
    n_k = itertools.accumulate((math.ceil(rate[i + 1] / rate[i]) for i in range(b, w)), operator.mul, initial=1)
    return max(n * memory[b + k] for k, n in enumerate(n_k))


def ser_block_cost(runtime: list[int], rate: list[int], data: list[int]) -> int:
    """
    Calculate running time of a subtree block with serialization.

    :param runtime: list of runtime values
    :param rate:    list of rate values
    :param data:    list of data values
    :return:        calculated cost
    """
    # return rate[b] * (data[b] + sum([(rate[i] / rate[b]) * runtime[i] for i in range(b, w + 1)]))
    return sum((r * t for r, t in zip(rate, runtime)), start=rate[0] * data[0])


def ser_block_latency(runtime: list[int], rate: list[int], data: list[int]) -> int:
    """
    Calculate relevant latency of a subtree block with serialization.

    :param runtime: list of runtime values
    :param rate:    list of rate values
    :param data:    list of data values
    :return:        calculated latency
    """
    return sum((n * t for n, t in zip(iser_mul_factor(rate), runtime)), start=data[0])


########################################################################################################################


def ser_block_submemory(memory: list[int], b: int, w: int) -> int:
    """
    Calculate cumulative memory of **chain block** [b, w] with serialization and data fetching/caching.

    :param memory:  list of memory values
    :param b:       barrier node
    :param w:       end node of block
    :return:        calculated memory value
    """
    return sum(itertools.islice(memory, b, w + 1))


def ser_block_subcost(runtime: list[int], rate: list[int], data: list[int], b: int, w: int) -> int:
    """
    Calculate running time of a **chain block** [b, w] with serialization and data fetching/caching.

    :param runtime: list of runtime values
    :param rate:    list of rate values
    :param data:    list of data values
    :param b:       barrier node
    :param w:       end node of block
    :return:        calculated cost
    """
    cost = sum((r * t for r, t in zip(itertools.islice(rate, b, w + 1),
                                      itertools.islice(runtime, b, w + 1))), start=rate[b] * data[b])
    return cost + rate[w + 1] * data[w + 1] if w < len(data) - 1 else cost


def ser_block_sublatency(runtime: list[int], rate: list[int], data: list[int], b: int, w: int, delay: int,
                         start: int, end: int) -> int:
    """
    Calculate relevant latency for **chain block** [b,w] with serialization and data fetching/caching.

    :param runtime: list of runtime values
    :param rate:    list of rate values
    :param data:    list of data values
    :param b:       barrier node
    :param w:       end node of block
    :param delay:   platform delay
    :param start:   fist node to consider
    :param end:     last node to consider
    :return:        calculated latency
    """
    if end < b or w < start:
        # Do not consider latency if there is no intersection
        return 0
    # Calculate data fetching and caching
    rw_data = [0] * len(range(b, w + 1))
    rw_data[0] += data[b]
    # w is no leaf
    rw_data[-1] += data[w + 1] if w != len(data) - 1 else 0
    e = min(w, end) + 1
    blk_lat = sum(n * (t + d) for i, n, t, d in zip(range(b, e),
                                                    iser_mul_factor(itertools.islice(rate, b, e)),
                                                    itertools.islice(runtime, b, e),
                                                    rw_data)
                  if start <= i)
    # Ignore delay if latency path starts within the subchain
    return delay + blk_lat if start < b else blk_lat


########################################################################################################################


def ser_subtree_memory(tree: nx.DiGraph, nodes: set[int]) -> int:
    """
    Calculate cumulative memory of a subtree.

    :param tree:    input tree
    :param nodes:   set of block nodes
    :return:        calculated memory
    """
    return sum(tree.nodes[v][MEMORY] for v in nodes)


def ser_subtree_cost(tree: nx.DiGraph, barr: int, nodes: set[int]) -> int:
    """
    Calculate running time of a **subtree** with serialization and data fetching/caching.

    :param tree:    input tree
    :param barr:    barrier node
    :param nodes:   set of block nodes
    :return:        calculated cost
    """
    p = next(tree.predecessors(barr))
    return sum((tree[next(tree.predecessors(v))][v][RATE] * tree.nodes[v][RUNTIME] +
                sum(vs[RATE] * vs[DATA] for s, vs in tree.succ[v].items() if s not in nodes)
                for v in nodes), start=tree[p][barr][RATE] * tree[p][barr][DATA])


def ser_pes_subchain_latency(tree: nx.DiGraph, barr: int, nodes: set[int], cpath: set[int]) -> int:
    """
    Calculate relevant latency of **chain** in group of **nodes** with serialization and **pessimistic** caching.

    :param tree:    input tree
    :param barr:    barrier node
    :param nodes:   set of block nodes
    :param cpath:   critical path nodes
    :return:        calculated latency
    """
    if not (subchain := sorted(nodes & cpath)):
        return 0
    p = next(tree.predecessors(barr))
    # Sum the instance number * runtime execution time + caching overhead of all block-egress edges
    return sum((n_v * (tree.nodes[v][RUNTIME] + sum(math.ceil(vs[RATE] / tree[u][v][RATE]) * vs[DATA]
                                                    for s, vs in tree.succ[v].items() if s not in nodes))
                for n_v, (u, v) in zip(iser_mul_factor(tree[i][j][RATE] for i, j in itertools.pairwise([p, *subchain])),
                                       itertools.pairwise([p, *subchain]))), start=tree[p][barr][DATA])


def ser_subchain_latency(tree: nx.DiGraph, barr: int, nodes: set[int], cpath: set[int]) -> int:
    """
    Calculate relevant latency of **chain** in group of **nodes** with serialization and data fetching/caching.

    :param tree:    input tree
    :param barr:    barrier node
    :param nodes:   set of block nodes
    :param cpath:   critical path nodes
    :return:        calculated latency
    """
    if barr not in cpath:
        return 0
    chain = sorted(nodes & cpath)
    # Get next cpath node following the given group
    cb = next(filter(lambda v: v in cpath, tree.successors(chain[-1])), None)
    p = next(tree.predecessors(barr))
    return sum(((n_v * tree.nodes[v][RUNTIME] +
                 (n_v * math.ceil(tree[v][cb][RATE] / tree[u][v][RATE]) * tree[v][cb][DATA]
                  if v == chain[-1] and cb is not None else 0))
                for n_v, (u, v) in zip(iser_mul_factor(tree[i][j][RATE] for i, j in itertools.pairwise([p, *chain])),
                                       itertools.pairwise([p, *chain]))), start=tree[p][barr][DATA])


########################################################################################################################


def par_inst_count(r_barr: int, r_v: int, N: int = 1) -> int:
    """
    Calculate instance number of a function considering the function/barrier rates and CPU count *N*.

    :param r_barr:  barrier node's ingress rate
    :param r_v:     call rate of node v
    :param N:       CPU count
    :return:        calculated instance count of node v
    """
    rel_r_v, sat_insts = r_v / r_barr, r_v % r_barr
    return sat_insts * math.ceil(math.ceil(rel_r_v) / N) + (r_barr - sat_insts) * math.ceil(math.floor(rel_r_v) / N)


def par_subtree_memory(tree: nx.DiGraph, barr: int, nodes: list[int] | set[int], N: int = 1) -> int:
    """
    Calculate memory demand of a subtree as the sum of cumulative and parallel execution components.

    :param tree:    input tree
    :param barr:    barrier node
    :param nodes:   set of block nodes
    :param N:       CPU count
    :return:        calculated memory value
    """
    r_b = tree[next(tree.predecessors(barr))][barr][RATE]
    return max(sum(tree.nodes[v][MEMORY] for v in nodes),
               max(min(math.ceil(tree[next(tree.predecessors(v))][v][RATE] / r_b), N) * tree.nodes[v][MEMORY]
                   for v in nodes))


def par_subtree_cost(tree: nx.DiGraph, barr: int, nodes: set[int], N: int = 1) -> int:
    """
    Calculate running time of a **subtree** with multiprocessing and data fetching/caching.

    :param tree:    input tree
    :param barr:    barrier node
    :param nodes:   set of block nodes
    :param N:       CPU count
    :return:        calculated cost
    """
    p = next(tree.predecessors(barr))
    r_b = tree[p][barr][RATE]
    return sum((par_inst_count(r_b, tree[next(tree.predecessors(v))][v][RATE], N) * tree.nodes[v][RUNTIME] +
                sum(par_inst_count(r_b, vs[RATE], N) * vs[DATA] for s, vs in tree.succ[v].items() if s not in nodes)
                for v in nodes), start=r_b * tree[p][barr][DATA])


def par_subchain_latency(tree: nx.DiGraph, barr: int, nodes: set[int], cpath: set[int], N: int = 1) -> int:
    """
    Calculate relevant latency of **chain** in group of **nodes** with serialization and data fetching/caching.

    :param tree:    input tree
    :param barr:    barrier node
    :param nodes:   set of block nodes
    :param cpath:   critical path nodes
    :param N:       CPU count
    :return:        calculated latency
    """
    if barr not in cpath:
        return 0
    chain = sorted(nodes & cpath)
    # Get next cpath node following the given group
    cb = next(filter(lambda v: v in cpath, tree.successors(chain[-1])), None)
    p = next(tree.predecessors(barr))
    return sum(((n_v * tree.nodes[v][RUNTIME] +
                 (n_v * math.ceil(tree[v][cb][RATE] / (tree[u][v][RATE] * N)) * tree[v][cb][DATA]
                  if v == chain[-1] and cb is not None else 0))
                for n_v, (u, v) in zip(ipar_mul_factor((tree[i][j][RATE]
                                                        for i, j in itertools.pairwise([p, *chain])), N),
                                       itertools.pairwise([p, *chain]))), start=tree[p][barr][DATA])


########################################################################################################################


def gen_subtree_memory(tree: nx.DiGraph, barr: int, nodes: set[int], N: int = 1) -> int:
    """
    Calculate memory demand of a subtree as the sum of cumulative and parallel execution components.

    :param tree:    input tree
    :param barr:    barrier node
    :param nodes:   set of block nodes
    :param N:       CPU count
    :return:        calculated memory value
    """
    r_b = tree[next(tree.predecessors(barr))][barr][RATE]
    return max(sum(tree.nodes[v][MEMORY] for v in nodes),
               max(min(math.ceil(tree[next(tree.predecessors(v))][v][RATE] / r_b),
                       # In general the max number of running instance can be bounded in threadpools
                       math.ceil(N / tree.nodes[v].get(CPU, 1))) * tree.nodes[v][MEMORY] for v in nodes))


def gen_subtree_cost(tree: nx.DiGraph, barr: int, nodes: set[int], N: int = 1,
                     exec_calc: collections.abc.Callable[[int, int, int], int] = lambda i, t, n: t) -> int:
    """
    Calculate running time of a **subtree** with multiprocessing and data fetching/caching while using *exec_calc*
    callable to recalculate function execution time based on the function's id (i), reference runtime (t) and available
    CPU cores (n).

    :param tree:        input tree
    :param barr:        barrier node
    :param nodes:       set of block nodes
    :param N:           CPU count
    :param exec_calc:   calculator function
    :return:            calculated cost
    """
    p = next(tree.predecessors(barr))
    r_b = tree[p][barr][RATE]
    return sum((par_inst_count(r_b, tree[next(tree.predecessors(v))][v][RATE], int(N / tree.nodes[v].get(CPU, 1))) *
                exec_calc(v, tree.nodes[v][RUNTIME], N) +
                sum(par_inst_count(r_b, vs[RATE], N) * vs[DATA] for s, vs in tree.succ[v].items() if s not in nodes)
                for v in nodes), start=r_b * tree[p][barr][DATA])


def gen_subchain_latency(tree: nx.DiGraph, barr: int, nodes: set[int], cpath: set[int], N: int = 1,
                         exec_calc: collections.abc.Callable[[int, int, int], int] = lambda i, t, n: t) -> int:
    """
    Calculate relevant latency of **chain** in group of **nodes** with serialization and data fetching/caching while
    using *exec_calc* callable to recalculate function execution time based on the function's id (i), reference runtime
    (t) and available CPU cores (n).

    :param tree:        input tree
    :param barr:        barrier node
    :param nodes:       set of block nodes
    :param cpath:       critical path nodes
    :param N:           CPU count
    :param exec_calc:   calculator function
    :return:            calculated latency
    """
    if barr not in cpath:
        return 0
    chain = sorted(nodes & cpath)
    # Get next cpath node following the given group
    cb = next(filter(lambda v: v in cpath, tree.successors(chain[-1])), None)
    p = next(tree.predecessors(barr))
    return sum(((n_v * exec_calc(v, tree.nodes[v][RUNTIME], int(N / tree.nodes[v].get(CPU, 1))) +
                 (n_v * math.ceil(tree[v][cb][RATE] / (tree[u][v][RATE] * N)) * tree[v][cb][DATA]
                  if v == chain[-1] and cb is not None else 0))
                for n_v, (u, v) in zip(igen_mul_factor(*zip(*((tree[i][j][RATE], int(N / tree.nodes[j].get(CPU, 1)))
                                                              for i, j in itertools.pairwise([p, *chain])))),
                                       itertools.pairwise([p, *chain]))), start=tree[p][barr][DATA])



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib
import pprint
import time

import cspy
import networkx as nx

from slambuc.alg import INFEASIBLE
from slambuc.alg.ext import *
from slambuc.alg.service.common import NAME, Flavor
from slambuc.alg.util import ibacktrack_chain
from slambuc.misc.generator import get_random_tree
from slambuc.misc.plot import draw_state_dag
from slambuc.misc.util import evaluate_gen_tree_partitioning


def test_par_csp_dag_creation(tree_file: str = pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml",
                              draw: bool = False, full: bool = False):
    tree = nx.read_gml(tree_file, destringizer=int)
    tree.graph[NAME] += "-par_csp"
    params = dict(tree=tree,
                  root=1,
                  flavors=[Flavor(6, 2, 1)],
                  # cp_end=10,
                  cpath=set(ibacktrack_chain(tree, 1, 10)),
                  # M=6,
                  # L=430,
                  # N=2,
                  delay=10)
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print('=' * 80)
    _s = time.perf_counter()
    models = [ret for ret in ibuild_gen_csp_dag(**params)]
    _d = time.perf_counter() - _s
    print(f"Model building time: sum: {_d * 1000} / avg: {_d * 1000 / len(models)} ms")
    for i, (dag, chains) in enumerate(models):
        print(f"Chain[{i}]:", chains)
        print(f"DAG[{i}]:", dag)
    if draw:
        for dag, chains in models:
            draw_state_dag(dag, chains, draw_weights=full)


def test_par_csp_solution(solver=cspy.BiDirectional, **kwargs):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-par_csp"
    params = dict(tree=tree,
                  root=1,
                  flavors=[Flavor(6, 2, 1)],
                  # cp_end=10,
                  cpath=set(ibacktrack_chain(tree, 1, 10)),
                  # M=6,
                  # L=430,
                  # N=2,
                  delay=10)
    L = 430
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print(f"Latency constraint: {L}")
    runtimes = []
    best_res = INFEASIBLE
    for i, (dag, chains) in enumerate(ibuild_gen_csp_dag(**params)):
        print(f" Run Bidirectional model[{i}] ".center(80, '='))
        print(f"Chain[{i}]:", chains)
        print(f"DAG[{i}]:", dag)
        max_res, min_res = [len(dag.edges), L], [1, 0]
        print(f"Alg params: {max_res=}, {min_res=}")
        _s = time.perf_counter()
        model = solver(dag, max_res, min_res, **kwargs)
        model.run()
        _d = time.perf_counter() - _s
        runtimes.append(_d)
        print(f"Exec. time: {_d * 1000} ms")
        cost, lat = model.total_cost, model.consumed_resources
        print(f"Result[{i}]:", model.path, "cost/lat:", cost, "/", lat)
        if cost is not None and cost < best_res[1]:
            best_res = [extract_grp_from_path(model.path), cost, lat[1]]
            print("New min cost partitioning found!")
    print('=' * 80)
    print(f"Sum time: {sum(runtimes) * 1000} ms")
    print(f"Best result: {best_res[0]} with cost/lat: {best_res[1]} / {best_res[2]}")


def evaluate_csp_model(file_name: str = pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", M: int = 6,
                       N: int = 2, L: int = math.inf):
    tree = nx.read_gml(file_name, destringizer=int)
    tree.graph[NAME] += "-par_csp"
    params = dict(tree=tree,
                  root=1,
                  # cp_end=10,
                  cpath=set(ibacktrack_chain(tree, 1, 10)),
                  # M=6,
                  # N=2,
                  flavors=[Flavor(M, N, 1)],
                  L=L,
                  delay=10)
    solvers = [cspy.BiDirectional,  # Bidirectional labelling algorithm with dynamic half-way point [(Tilk at al. 2017]
               cspy.Tabu,  # Simple Tabu-esque algorithm
               cspy.GreedyElim,  # Simple Greedy elimination (eliminates resource infeasible edges iteratively)
               cspy.GRASP,  # Greedy Randomised Adaptive Search Procedure [Ferone et al. 2019]
               # cspy.PSOLGENT # Particle Swarm Optimization [Marinakis et al. 2017] - bugs in swarm initialization
               ]
    runtimes = []
    results = []
    for solver in solvers:
        print(f" {solver.__name__} solver  ".center(80, '='))
        _s = time.perf_counter()
        partition, opt_cost, opt_lat = csp_gen_tree_partitioning(solver=solver, **params)
        runtimes.append(time.perf_counter() - _s)
        results.append((partition, opt_cost, opt_lat))
        evaluate_gen_tree_partitioning(tree, partition, opt_cost, opt_lat, params['root'], params['flavors'],
                                       params['cp_end'], params['L'], params['delay'])
    for solver, rt, (partition, opt_cost, opt_lat) in zip(solvers, runtimes, results):
        print(f"[{solver.__name__}] Exec. time: {rt * 1000} ms")
        print(f"[{solver.__name__}] Partitioning: {partition}, {opt_cost = }, {opt_lat = }")


def test_gen_csp_dag_creation(tree_file: str = pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml",
                              draw: bool = False, full: bool = False):
    tree = nx.read_gml(tree_file, destringizer=int)
    tree.graph[NAME] += "-par_csp"
    params = dict(tree=tree,
                  root=1,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  # cp_end=10,
                  cpath=set(ibacktrack_chain(tree, 1, 10)),
                  # L=430,
                  delay=10)
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print('=' * 80)
    _s = time.perf_counter()
    models = [ret for ret in ibuild_gen_csp_dag(**params)]
    _d = time.perf_counter() - _s
    print(f"Model building time: sum: {_d * 1000} / avg: {_d * 1000 / len(models)} ms")
    for i, (dag, chains) in enumerate(models):
        print(f"Chain[{i}]:", chains)
        print(f"DAG[{i}]:", dag)
    if draw:
        for dag, chains in models:
            draw_state_dag(dag, chains, draw_weights=full)


def test_gen_csp_solution(solver=cspy.BiDirectional, **kwargs):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-par_csp"
    params = dict(tree=tree,
                  root=1,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  # cp_end=10,
                  cpath=set(ibacktrack_chain(tree, 1, 10)),
                  # L=430,
                  delay=10)
    L = 430
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print(f"Latency constraint: {L}")
    runtimes = []
    best_res = INFEASIBLE
    for i, (dag, chains) in enumerate(ibuild_gen_csp_dag(**params)):
        print(f" Run Bidirectional model[{i}] ".center(80, '='))
        print(f"Chain[{i}]:", chains)
        print(f"DAG[{i}]:", dag)
        max_res, min_res = [len(dag.edges), L], [1, 0]
        print(f"Alg params: {max_res=}, {min_res=}")
        _s = time.perf_counter()
        model = solver(dag, max_res, min_res, **kwargs)
        model.run()
        _d = time.perf_counter() - _s
        runtimes.append(_d)
        print(f"Exec. time: {_d * 1000} ms")
        cost, lat = model.total_cost, model.consumed_resources
        print(f"Result[{i}]:", model.path, "cost/lat:", cost, "/", lat)
        if cost is not None and cost < best_res[1]:
            best_res = [extract_grp_from_path(model.path), cost, lat[1]]
            print("New min cost partitioning found!")
    print('=' * 80)
    print(f"Sum time: {sum(runtimes) * 1000} ms")
    print(f"Best result: {best_res[0]} with cost/lat: {best_res[1]} / {best_res[2]}")


########################################################################################################################


def run_test(tree: nx.DiGraph, root: int, flavors: list[Flavor], cp_end: int, L: int, delay: int):
    _s = time.perf_counter()
    partition, opt_cost, opt_lat = csp_gen_tree_partitioning(tree, root, flavors, L=L, cp_end=cp_end, delay=delay)
    _d = time.perf_counter() - _s
    print(f"Sum time: {_d * 1000} ms")
    evaluate_gen_tree_partitioning(tree, partition, opt_cost, opt_lat, root, flavors, cp_end, L, delay)
    return partition, opt_cost, opt_lat


def test_par_tree():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-par_csp"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  flavors=[Flavor(6, 2, 1)],
                  # L = math.inf
                  L=430,
                  delay=10)
    run_test(**params)


def test_random_par_tree(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-par_csp"
    params = dict(tree=tree,
                  root=1,
                  flavors=[Flavor(6, 2, 1)],
                  cp_end=n,
                  L=math.inf,
                  delay=10)
    run_test(**params)


def test_gen_tree():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-par_csp"
    params = dict(tree=tree,
                  root=1,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  cp_end=10,
                  # L = math.inf
                  L=430,
                  delay=10)
    run_test(**params)


def test_random_gen_tree(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-par_csp"
    params = dict(tree=tree,
                  root=1,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  cp_end=n,
                  L=math.inf,
                  delay=10)
    run_test(**params)


if __name__ == '__main__':
    # test_par_csp_dag_creation(draw=True, full=True)
    # test_par_csp_dag_creation(draw=False)
    # test_par_csp_solution(solver=cspy.BiDirectional, direction="forward")
    # test_par_csp_solution(solver=cspy.PSOLGENT)
    # evaluate_csp_model(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", M=6, N=1, L=430)
    # evaluate_csp_model(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", M=6, N=2, L=400)
    #
    # test_gen_csp_dag_creation(draw=False)
    # test_gen_csp_dag_creation(draw=True, full=True)
    # test_gen_csp_solution(solver=cspy.BiDirectional, direction="forward")
    #
    # test_par_tree()
    test_random_par_tree(n=30)
    #
    # test_gen_tree()
    # test_random_gen_tree(n=10)



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import heapq
import itertools
import math

import networkx as nx

from slambuc.alg import INFEASIBLE, T_BLOCK, T_RESULTS
from slambuc.alg.service import MEMORY, RATE, DATA
from slambuc.alg.util import ibacktrack_chain, recalculate_partitioning, par_subchain_latency, par_subtree_memory

# Naming convention for state-space DAG
START, END = 's', 't'


def get_bounded_greedy_block(tree: nx.DiGraph, root: int, M: int, N: int = 1, cp_end: int = None,
                             cp_cuts: set[int] = frozenset()) -> tuple[T_BLOCK, list[int]]:
    """
    Calculate a partition block based on the memory limit *M* by iteratively merging edges with the largest weights
    started from the given *root*.

    Filter out mandatory cuts of *cp_cuts* on the cpath form merging, while merges other cpath edges.

    :param tree:    service graph annotated with node runtime(ms), edge rate and edge data unit size
    :param root:    root node of the tree
    :param M:       upper memory bound of the partition blocks in MB
    :param N:       available CPU core count
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param cp_cuts: barrier nodes of mandatory cuts on the critical path
    :return:        calculated partition block and the separated neighbouring nodes
    """
    # Init
    blk, blk_mem, cp_set, neighs = {root}, tree.nodes[root][MEMORY], set(ibacktrack_chain(tree, root, cp_end)), []

    def add_neighbours(node: int):
        """Add neighbours of the given *node* while keeping the order with a heap queue"""
        for sn, d in tree[node].items():
            if sn in cp_cuts:
                continue
            # Set edge weight to -inf for cpath nodes not included in precalculated cuts for designated must-merge
            heapq.heappush(neighs, (-1 * (d[RATE] * d[DATA] if sn not in cp_set else math.inf), sn))

    # Collect possible cut/merge edges based on the neighbouring nodes
    # Set edge weight to inf for cpath nodes not included in precalculated cuts for designated merging
    add_neighbours(root)
    while neighs:
        # Chose the largest data transfer edge for merging -> minimize cut weights
        _, v = heapq.heappop(neighs)
        # Stop extension of the current block due to memory limit
        if par_subtree_memory(tree, root, blk | {v}, N) > M:
            break
        blk.add(v)
        add_neighbours(v)
    # Collect root nodes of cut subtrees by the given block
    succ = [sn for n in blk for sn in tree[n] if sn not in blk]
    return sorted(blk), succ


def min_weight_greedy_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, N: int = 1,
                                   delay: int = 1, metrics: bool = True, **kwargs) -> T_RESULTS:
    """
    Calculates memory-bounded tree partitioning in a greedy manner without any latency limit.

    :param tree:    service graph annotated with node runtime(ms), edge rate and edge data unit size
    :param root:    root node of the tree
    :param M:       upper memory bound of the partition blocks in MB
    :param N:       available CPU core count
    :param delay:   invocation delay between blocks
    :param metrics: return calculated sum cost and critical path latency
    :return:        tuple of derived partitioning, sum cost, and the latency on the critical path (root, cp_end)
    """
    partition, succ = [], collections.deque((root,))
    # Iteratively grow partition blocks starting from root and restarting it at the neighbouring nodes
    while succ:
        blk, next_succ = get_bounded_greedy_block(tree, succ.popleft(), M, N)
        succ.extend(next_succ)
        partition.append(blk)
    sum_cost, sum_lat = recalculate_partitioning(tree, partition, root, N, None, delay) if metrics else (None, None)
    return partition, sum_cost, sum_lat


########################################################################################################################


def get_feasible_cpath_split(tree: nx.DiGraph, root: int, cp_end: int, M: int, L: int, N: int = 1,
                             delay: int = 1) -> set[int]:
    """
    Calculate feasible splitting of the critical path that meets given memory *M* and latency *L* limits.

    :param tree:    service graph annotated with node runtime(ms), edge rate and edge data unit size
    :param root:    root node of the tree
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param M:       upper memory bound of the partition blocks in MB
    :param L:       latency limit defined on the critical path in ms
    :param N:       available CPU core count
    :param delay:   invocation delay between blocks
    :return:        set of barrier nodes of calculated critical path blocks
    """
    cpath = list(reversed(list(ibacktrack_chain(tree, root, cp_end))))
    cp_len, cp_set = len(cpath), set(cpath)
    # Check trivial single block subcase
    if par_subchain_latency(tree, root, cp_set, cp_set, N) <= L and par_subtree_memory(tree, root, cpath, N) <= M:
        return {root}
    cuts, best_feasible_lat, best_feasible_cut = set(), math.inf, None
    while True:
        best_lat, best_cut = math.inf, None
        # Iterate over possible cut
        for c in range(1, cp_len):
            if c in cuts:
                continue
            # Calculate relevant metrics
            blk_lats, blk_mems = zip(*((par_subchain_latency(tree, cpath[i], set(cpath[i:j]), cp_set, N),
                                        par_subtree_memory(tree, cpath[i], cpath[i:j], N))
                                       for i, j in itertools.pairwise([0, *sorted(cuts | {c}), cp_len])))
            # Store better case for the new cut *c* added to prior chosen *cuts*
            if (sum_lats := sum(blk_lats) + (len(blk_lats) - 1) * delay) < best_lat:
                best_lat, best_cut = sum_lats, c
            # Track feasible splitting regarding the block memories
            if all(map(lambda _m: _m <= M, blk_mems)) and sum_lats < best_feasible_lat:
                best_feasible_lat, best_feasible_cut = sum_lats, c
        # Return the first feasible cuts
        if best_feasible_lat <= L:
            cuts |= {best_feasible_cut}
            break
        if best_cut is None:
            # No latency-limited subcase can be achieved or all iterative cut subcases are examined
            return None
        else:
            # Greedily select the locally best cut (the largest reduce in lat possibly but not necessarily in opt cuts)
            cuts.add(best_cut)
    return {root, *(cpath[c] for c in cuts)}


def get_min_cpath_split(tree: nx.DiGraph, root: int, cp_end: int, M: int, L: int, N: int = 1,
                        delay: int = 1) -> set[int]:
    """
    Calculate min-latency splitting of the critical path that meets given memory *M* and latency *L* limits.

    :param tree:    service graph annotated with node runtime(ms), edge rate and edge data unit size
    :param root:    root node of the tree
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param M:       upper memory bound of the partition blocks in MB
    :param L:       latency limit defined on the critical path in ms
    :param N:       available CPU core count
    :param delay:   invocation delay between blocks
    :return:        set of barrier nodes of calculated critical path blocks
    """
    cpath = list(reversed(list(ibacktrack_chain(tree, root, cp_end))))
    cp_set = set(cpath)
    # Check trivial single block subcase
    if par_subchain_latency(tree, root, cp_set, cp_set, N) <= L and par_subtree_memory(tree, root, cpath, N) <= M:
        return {root}
    # Initiate data structure for DAG
    _cache = collections.defaultdict(list)
    _cache.update({START: [START], END: [END]})
    dag = nx.DiGraph(directed=True, **tree.graph)
    for i, (prev, b) in enumerate(itertools.pairwise(itertools.chain([START], cpath))):
        blk = set()
        for w in cpath[i:]:
            blk.add(w)
            # Skip infeasible subcase due to memory constraint
            if par_subtree_memory(tree, b, blk, N) > M:
                break
            blk_id = f"{b}|{w}"
            _cache[w].append(blk_id)
            blk_lat = par_subchain_latency(tree, b, blk, cp_set, N)
            # Add invocation delay for inter-block invocations
            if b != root and blk_lat > 0:
                blk_lat += delay
            # Add connection between related subcases
            for p in _cache[prev]:
                dag.add_edge(p, blk_id, weight=blk_lat)
    # Add connection for ending subcases
    for p in _cache[cpath[-1]]:
        dag.add_edge(p, END, weight=0)
    min_lat, sp = nx.single_source_dijkstra(dag, source=START, target=END)
    return {int(v.split('|')[0]) for v in sp[1:-1]} if min_lat <= L else None


def min_weight_partition_heuristic(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                                   N: int = 1, cp_end: int = None, delay: int = 1, metrics: bool = True) -> T_RESULTS:
    """
    Greedy heuristic algorithm to calculate partitioning of the given *tree* regarding the given memory *M* and
    latency *L* limits.
    It uses a greedy approach to calculate a low-cost critical path cut (might miss feasible solutions).
    It may conclude the partitioning problem infeasible despite there exist one with large costs.

    :param tree:    service graph annotated with node runtime(ms), edge rate and edge data unit size
    :param root:    root node of the tree
    :param M:       upper memory bound of the partition blocks in MB
    :param L:       latency limit defined on the critical path in ms
    :param N:       available CPU core count
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :param metrics: return calculated sum cost and critical path latency
    :return:        tuple of derived partitioning, sum cost, and the latency on the critical path (root, cp_end)
    """
    # Greedy search for feasible cut of the critical path
    cpath_cuts = get_feasible_cpath_split(tree, root, cp_end, M, L, N, delay)
    # No feasible cuts are found
    if cpath_cuts is None:
        return INFEASIBLE
    partition, succ = [], collections.deque((root,))
    # Iteratively grow partition blocks starting from root and restarting it at the neighbouring nodes
    while succ:
        v = succ.popleft()
        blk, next_succ = get_bounded_greedy_block(tree, v, M, N, cp_end, cpath_cuts)
        succ.extend(next_succ)
        partition.append(blk)
    sum_cost, sum_lat = recalculate_partitioning(tree, partition, root, N, cp_end, delay) if metrics else (None, None)
    return partition, sum_cost, sum_lat


def min_lat_partition_heuristic(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                                N: int = 1, cp_end: int = None, delay: int = 1, metrics: bool = True) -> T_RESULTS:
    """
    Greedy heuristic algorithm to calculate partitioning of the given *tree* regarding the given memory *M* and
    latency *L* limits.
    It uses Dijkstra's algorithm to calculate the critical path cut with the lowest latency (might be expensive).
    It always returns a latency-feasible solution if it exists.

    :param tree:    service graph annotated with node runtime(ms), edge rate and edge data unit size
    :param root:    root node of the tree
    :param M:       upper memory bound of the partition blocks in MB
    :param L:       latency limit defined on the critical path in ms
    :param N:       available CPU core count
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :param metrics: return calculated sum cost and critical path latency
    :return:        tuple of derived partitioning, sum cost, and the latency on the critical path (root, cp_end)
    """
    # Greedy search for min-latency cut of the critical path
    cpath_cuts = get_min_cpath_split(tree, root, cp_end, M, L, N, delay)
    # No feasible cuts are found
    if cpath_cuts is None:
        return INFEASIBLE
    partition, succ = [], collections.deque((root,))
    # Iteratively grow partition blocks starting from root and restarting it at the neighbouring nodes
    while succ:
        v = succ.popleft()
        blk, next_succ = get_bounded_greedy_block(tree, v, M, N, cp_end, cpath_cuts)
        succ.extend(next_succ)
        partition.append(blk)
    sum_cost, sum_lat = recalculate_partitioning(tree, partition, root, N, cp_end, delay) if metrics else (None, None)
    return partition, sum_cost, sum_lat



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import itertools
import math
import operator

import networkx as nx

from slambuc.alg import INFEASIBLE, T_RESULTS
from slambuc.alg.service import RATE, DATA, PLATFORM
from slambuc.alg.util import recreate_subtree_blocks, recalculate_partitioning


def min_weight_subchain_split(tree: nx.DiGraph, root: int) -> set[int]:
    """
    Return chain-based edge cuts with the minimal edge weight (amount of transferred data).

    The splitting marks the edge with the largest weight at each branching nodes to be a must-merge edge.

    :param tree:    service graph annotated with node runtime(ms), edge rate and edge data unit size
    :param root:    root node of the tree
    :return:        set of barrier nodes
    """
    return {root}.union(*(set(tree.successors(v)) -
                          {max(tree.successors(v), key=lambda s: tree[v][s][RATE] * tree[v][s][DATA])}
                          for v in tree if v is not PLATFORM and len(tree.succ[v]) > 1))


def min_weight_chain_decomposition(tree: nx.DiGraph, root: int, N: int = 1, cp_end: int = None,
                                   delay: int = 1, metrics: bool = True, **kwargs) -> T_RESULTS:
    """
    Minimal edge-weight chain-based tree partitioning (O(n)) without memory and latency constraints.

    Although latency is not considered on the critical path the algorithm reports it with the sum cost.

    :param tree:    service graph annotated with node runtime(ms) and edge rate
    :param root:    root node of the tree
    :param N:       available CPU core count
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :param metrics: return calculated sum cost and critical path latency
    :return:        tuple of derived partitioning, sum cost, and the latency on the critical path (root, cp_end)
    """
    partition = recreate_subtree_blocks(tree, barr=min_weight_subchain_split(tree, root))
    sum_cost, sum_lat = recalculate_partitioning(tree, partition, root, N, cp_end, delay) if metrics else (None,) * 2
    return partition, sum_cost, sum_lat


########################################################################################################################


def min_weight_ksplit(tree: nx.DiGraph, root: int, k: int) -> set[int]:
    """
    Minimal data-transfer tree clustering into *k* clusters with k-1 cuts without memory and latency constraints.

    The clustering algorithm is based on the maximum split clustering algorithm(O(n^3)) which ranks the edges (paths)
    based on the amount of transferred data.

    Details in: M. Maravalle et al.: “Clustering on trees,” Computational Statistics & Data Analysis, vol. 24, no. 2,
    pp. 217–234, Apr. 1997, doi: 10.1016/S0167-9473(96)00062-X.

    :param tree:    service graph annotated with node runtime(ms), edge rate and edge data unit size
    :param root:    root node of the tree
    :param k:       number of clusters
    :return:        set of barrier nodes
    """
    dist_tree = nx.to_undirected(tree)
    # Define distance of two nodes as the reciprocal of the sum transferred data between the nodes
    D = {(u, v): sum(1 / (attr[RATE] * attr[DATA]) if attr[RATE] and attr[DATA] else 0
                     for i, j, attr in tree.edges(next(nx.all_simple_paths(dist_tree, u, v)), data=True))
         for u, v in itertools.combinations(filter(lambda n: n is not PLATFORM, tree), 2)}
    edges, rank = set(tree.edges(filter(lambda n: n is not PLATFORM, tree))), 1
    labeled = collections.deque(maxlen=k - 1)
    # Iterate paths from the min distant element
    for (i, j), _ in sorted(list(D.items()), key=operator.itemgetter(1)):
        path_edges = set(e for e in tree.edges(next(nx.all_simple_paths(dist_tree, i, j))))
        # If there is unlabelled edge on the given path
        if unlabeled := path_edges & edges:
            labeled.extend((rank, b) for _, b in unlabeled)
            edges -= unlabeled
            # If all edges are labeled -> stop
            if not edges:
                break
            rank += 1
    return {root}.union(b for _, b in labeled)


def min_weight_ksplit_clustering(tree: nx.DiGraph, root: int, k: int = None, N: int = 1, cp_end: int = None,
                                 delay: int = 1, metrics: bool = True, **kwargs) -> T_RESULTS:
    """
    Minimal data-transfer tree clustering into *k* clusters (with k-1 cuts) without memory and latency constraints.

    Although latency is not considered on the critical path the algorithm reports it along with the sum cost.

    :param tree:    service graph annotated with node runtime(ms), edge rate and edge data unit size
    :param root:    root node of the tree
    :param k:       number of clusters
    :param N:       available CPU core count
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :param metrics: return calculated sum cost and critical path latency
    :return:        tuple of derived partitioning, sum cost, and the latency on the critical path (root, cp_end)
    """
    k = k if k is not None else math.ceil(math.sqrt(len(tree) - 1))
    partition = recreate_subtree_blocks(tree, barr=min_weight_ksplit(tree, root, k))
    sum_cost, sum_lat = recalculate_partitioning(tree, partition, root, N, cp_end, delay) if metrics else (None, None)
    return partition, sum_cost, sum_lat


def min_weight_tree_clustering(tree: nx.DiGraph, root: int, L: int = math.inf, N: int = 1, cp_end: int = None,
                               delay: int = 1, metrics: bool = True, **kwargs) -> T_RESULTS:
    """
    Minimal data-transfer tree clustering into without memory constraints.

    Iteratively calculates *k-1* different ksplit clustering in reverse order until an L-feasible solution is found.

    Although latency is not considered on the critical path the algorithm reports it with the sum cost.

    :param tree:    service graph annotated with node runtime(ms), edge rate and edge data unit size
    :param root:    root node of the tree
    :param L:       latency limit defined on the critical path in ms
    :param N:       available CPU core count
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :param metrics: return calculated sum cost and critical path latency
    :return:        tuple of derived partitioning, sum cost, and the latency on the critical path (root, cp_end)
    """
    best_result = INFEASIBLE
    for k in reversed(range(1, len(tree) + 1)):
        partition = recreate_subtree_blocks(tree, barr=min_weight_ksplit(tree, root, k))
        cost, lat = recalculate_partitioning(tree, partition, root, N, cp_end, delay) if metrics else (math.inf,) * 2
        if cost <= best_result[1] and lat <= L:
            best_result = partition, cost, lat
    return best_result



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import itertools
import math
from collections.abc import Generator

import networkx as nx
import pulp as lp

from slambuc.alg import LP_LAT, INFEASIBLE, T_RESULTS, T_PART
from slambuc.alg.service import *
from slambuc.alg.util import (ipowerset, ipostorder_dfs, ibacktrack_chain, recreate_subtree_blocks, induced_subtrees,
                              ser_subtree_cost, ser_subchain_latency, ser_subtree_memory, verify_limits, x_eval)


def ifeasible_greedy_subtrees(tree: nx.DiGraph, root: int, M: int) -> Generator[tuple[int, set[int]]]:
    """
    Generate feasible subtrees in a combinatorial way, which meet the connectivity and memory constraint *M*.
    
    Block metrics are calculated based on serialized execution platform model.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :return:        generator of subtree root and regarding subtree nodes
    """
    for st in ipowerset(tuple(filter(lambda n: n is not PLATFORM, tree)), start=1):
        st = set(st)
        for v in itertools.islice(sorted(st, reverse=True), len(st) - 1):
            if next(tree.predecessors(v)) not in st:
                break
        else:
            if ser_subtree_memory(tree, st) <= M:
                yield min(st), st


def ifeasible_subtrees(tree: nx.DiGraph, root: int, M: int, filtered: bool = True) -> Generator[tuple[int, set[int]]]:
    """
    Generate M-feasible(connected) subtrees and roots in a bottom-up way, which meet the memory constraint *M*.

    Block metrics are calculated based on serialized execution platform model.

    :param tree:        service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:        root node of the graph
    :param M:           upper memory bound of the partition blocks (in MB)
    :param filtered:    filter our infeasible subtrees
    :return:            generator of subtree root and regarding subtree nodes
    """
    subtrees = collections.defaultdict(list)
    for _, n in ipostorder_dfs(tree, root):
        for children in ipowerset(tuple(tree.successors(n))):
            for sts in itertools.product(*(subtrees[c] for c in children)):
                st = {n}.union(*sts)
                if not filtered or ser_subtree_memory(tree, st) <= M:
                    subtrees[n].append(st)
                    yield n, st
        for c in tree.succ[n]:
            del subtrees[c]


def build_tree_cfg_model(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                         cpath: set[int] = frozenset(), delay: int = 1,
                         isubtrees: iter = ifeasible_subtrees) -> tuple[lp.LpProblem, dict[int, list[lp.LpVariable]]]:
    """
    Generate the configuration ILP model using serialized metric calculation.

    :return: tuple of the created model and list of decision variables
    """
    # Model
    model = lp.LpProblem(name="Tree_Partitioning", sense=lp.LpMinimize)
    # Decision variables with precalculated coefficients
    c_x, l_x, X_n = [], [], collections.defaultdict(list)
    for i, (b, nodes) in enumerate(isubtrees(tree, root, M)):
        # Decision variable for the subtree
        x = lp.LpVariable(f"x_{b:02d}_{i}", cat=lp.LpBinary)
        # Cache block for faster partition recreation
        x.blk = sorted(nodes)
        # Add subtree block cost
        c_x.append(ser_subtree_cost(tree, b, nodes) * x)
        # Add subtree block latency if required
        if b in cpath:
            st_lat = ser_subchain_latency(tree, b, nodes, cpath)
            l_x.append((st_lat + delay) * x if b != root else st_lat * x)
        # Cache node coverage of subtree block
        for n in nodes:
            X_n[n].append(x)
    # Objective
    model += lp.lpSum(c_x)
    # Feasibility constraints
    for i, x_i in X_n.items():
        model += lp.lpSum(x_i) == 1, f"Cf_{i:03d}"
    # Latency constraint
    model += lp.lpSum(l_x) <= L if L < math.inf else lp.lpSum(l_x) >= 0, LP_LAT
    return model, X_n


def tree_cfg_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                          cp_end: int = None, delay: int = 1, solver: lp.LpSolver = None,
                          timeout: int = None, **lpargs) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a tree based on configuration LP formulation and greedy subcase
    generation.
    
    Block metrics are calculated based on serialized execution platform model.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :param L:       latency limit defined on the critical path (in ms)
    :param cp_end:  tail node of the critical path in the form of subchain[root -> c_pend]
    :param delay:   invocation delay between blocks
    :param solver:  specific solver class (default: COIN-OR CBC)
    :param timeout: time limit in sec
    :param lpargs:  additional LP solver parameters
    :return:        tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    model, X = build_tree_cfg_model(tree, root, M, L, cpath, delay, isubtrees=ifeasible_greedy_subtrees)
    status = model.solve(solver=solver if solver else lp.PULP_CBC_CMD(mip=True, msg=False, timeLimit=timeout, **lpargs))
    if status == lp.LpStatusOptimal:
        opt_cost, opt_lat = lp.value(model.objective), lp.value(model.constraints[LP_LAT])
        return recreate_subtrees_from_xdict(tree, X), opt_cost, L + opt_lat if L < math.inf else opt_lat
    else:
        return INFEASIBLE


def tree_hybrid_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                             cp_end: int = None, delay: int = 1, solver: lp.LpSolver = None,
                             timeout: int = None, **lpargs) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a tree based on configuration LP formulation and hybrid subcase
    generation.

    Block metrics are calculated based on serialized execution platform model.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :param L:       latency limit defined on the critical path (in ms)
    :param cp_end:  tail node of the critical path in the form of subchain[root -> c_pend]
    :param delay:   invocation delay between blocks
    :param solver:  specific solver class (default: COIN-OR CBC)
    :param timeout: time limit in sec
    :param lpargs:  additional LP solver parameters
    :return:        tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    model, X = build_tree_cfg_model(tree, root, M, L, cpath, delay)
    status = model.solve(solver=solver if solver else lp.PULP_CBC_CMD(mip=True, msg=False, timeLimit=timeout, **lpargs))
    if status == lp.LpStatusOptimal:
        opt_cost, opt_lat = lp.value(model.objective), lp.value(model.constraints[LP_LAT])
        return recreate_subtrees_from_xdict(tree, X), opt_cost, L + opt_lat if L < math.inf else opt_lat
    else:
        return INFEASIBLE


def extract_subtrees_from_xdict(model: lp.LpProblem) -> T_PART:
    """
    Recreate partitioning blocks from metadata cached in variables objects.

    :param model:   LP problem model
    :return:        calculated partitioning
    """
    return sorted(x.blk for x in filter(lambda x: round(x.varValue) == 1, model.variables()))


def recreate_subtrees_from_xdict(tree: nx.DiGraph, Xn: dict[int, list[lp.LpVariable]]) -> T_PART:
    """
    Extract barrier nodes from variable names (x_{b}_{w}) and recreate partitioning blocks.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param Xn:      specific structure of decision variables
    :return:        calculated partitioning
    """
    barr = set(int(next(filter(lambda x: round(x.varValue) == 1, x_n)).name.split('_', 2)[1]) for x_n in Xn.values())
    return recreate_subtree_blocks(tree=tree, barr=barr)


########################################################################################################################


def build_greedy_tree_mtx_model(tree: nx.DiGraph, root: int = 1, M: int = math.inf,
                                L: int = math.inf, cpath: set[int] = frozenset(), subchains: bool = False,
                                delay: int = 1) -> tuple[lp.LpProblem, dict[int, dict[int, lp.LpVariable]]]:
    """
    Generate the matrix ILP model in a greedy manner.

    Block metrics are calculated based on serialized execution platform model.

    :return: tuple of the created model and list of decision variables
    """
    # Model
    model = lp.LpProblem(name="Tree_Partitioning", sense=lp.LpMinimize)
    # Empty decision variable matrix
    X = {j: {} for j in filter(lambda n: n is not PLATFORM, tree)}
    # Objective
    sum_cost = lp.LpAffineExpression()
    for j in X:
        cost_pre = 0
        nodes = set()
        for v in nx.dfs_preorder_nodes(tree, source=j):
            nodes |= {v}
            cost_vj = ser_subtree_cost(tree, j, nodes)
            X[v][j] = lp.LpVariable(f'x_{v:02d}_{j:02d}', cat=lp.LpBinary) if v != root else 1
            sum_cost += (cost_vj - cost_pre) * X[v][j]
            cost_pre = cost_vj
    model += sum_cost
    # Feasibility constraints
    for i in X:
        model += lp.lpSum(X[i]) == 1, f"Cf_{i:02d}"
    # Knapsack constraints
    for j in X:
        model += lp.lpSum(tree.nodes[i][MEMORY] * X[i][j]
                          for i in nx.dfs_preorder_nodes(tree, source=j)) <= M, f"Ck_{j:02d}"
    # Connectivity constraints
    for j in X:
        for u, v in nx.dfs_edges(tree, source=j):
            model += X[u][j] - X[v][j] >= 0, f"Cc_{j:02d}_{u:02d}_{v:02d}"
    # Path-tree constraints
    if subchains:
        for j in X:
            for v in nx.dfs_preorder_nodes(tree, source=j):
                if tree.succ[v]:
                    model += lp.lpSum(X[i][j] for i in tree.successors(v)) <= 1, f"Cp_{j:02d}_{v:02d}"
    # Latency constraint
    sum_lat = lp.LpAffineExpression()
    for j in X:
        if j not in cpath:
            continue
        lat_pre = 0
        nodes = set()
        for v in nx.dfs_preorder_nodes(tree, source=j):
            nodes |= {v}
            # Add subtree block latency if required
            lat_vj = ser_subchain_latency(tree, j, nodes, cpath)
            sum_lat += (lat_vj - lat_pre) * X[v][j]
            if v == j and j != root:
                sum_lat += delay * X[v][j]
            lat_pre = lat_vj
    if L < math.inf:
        model += sum_lat <= L, LP_LAT
    else:
        # Add redundant constraint to implicitly calculate the latency value
        model += sum_lat >= 0, LP_LAT
    return model, X


def build_tree_mtx_model(tree: nx.DiGraph, root: int = 1, M: int = math.inf,
                         L: int = math.inf, cpath: set[int] = frozenset(), subchains: bool = False,
                         delay: int = 1) -> tuple[lp.LpProblem, dict[int, dict[int, lp.LpVariable]]]:
    """
    Generate the matrix ILP model directly from formulas.
    
    Block metrics are calculated based on serialized execution platform model.

    :return: tuple of the created model and list of decision variables
    """
    # Model
    model = lp.LpProblem(name="Tree_Partitioning", sense=lp.LpMinimize)
    # Decision variable matrix with trivial variables
    X = {j: {j: lp.LpVariable(f"x_{j:02d}_{j:02d}", cat=lp.LpBinary) if j != root else 1}
         for j in tree.nodes if j is not PLATFORM}
    # Empty objective
    sum_cost = lp.LpAffineExpression()
    # Empty latency constraint
    sum_lat = lp.LpAffineExpression()
    # Generate decision variables
    for (p, j), st_edges in induced_subtrees(tree, root):
        blk_mem = tree.nodes[j][MEMORY] * X[j][j]
        # Add coefficients for single node block [j]
        r_j, d_j, t_j = tree[p][j][RATE], tree[p][j][DATA], tree.nodes[j][RUNTIME]
        sum_cost += (r_j * (d_j + t_j) + sum(js[RATE] * js[DATA] for js in tree.succ[j].values())) * X[j][j]
        if j in cpath:
            jc = next(filter(lambda c: c in cpath, tree.successors(j)), None)
            sum_lat += (d_j + t_j + (math.ceil(tree[j][jc][RATE] / r_j) * tree[j][jc][DATA] if jc else 0)) * X[j][j]
            if j != root:
                sum_lat += delay * X[j][j]
        # Cache instance factor for nodes in cpath
        n_v = 1
        # Candidate nodes for block_j
        for u, v in st_edges:
            X[v][j] = lp.LpVariable(f"x_{v:02d}_{j:02d}", cat=lp.LpBinary)
            blk_mem += tree.nodes[v][MEMORY] * X[v][j]
            # Add coefficients for merging single node v to block [j,...]
            r_v, d_v, t_v = tree[u][v][RATE], tree[u][v][DATA], tree.nodes[v][RUNTIME]
            sum_cost += (r_v * (t_v - d_v) + sum(vs[RATE] * vs[DATA] for vs in tree.succ[v].values())) * X[v][j]
            # u -> v edge on cpath
            if v in cpath:
                n_v *= math.ceil(r_v / tree[next(tree.predecessors(u))][u][RATE])
                vc = next(filter(lambda c: c in cpath, tree.successors(v)), None)
                w_v = math.ceil(tree[v][vc][RATE] / r_v) * tree[v][vc][DATA] if vc else 0
                sum_lat += n_v * (t_v - d_v + w_v) * X[v][j]
            # Connectivity constraint
            model += X[u][j] - X[v][j] >= 0, f"Cc_{j:02d}_{u:02d}_{v:02d}"
        # Knapsack constraint, X[l][l] <= M for each leaf node l can be omitted
        if len(blk_mem) > 1:
            model += blk_mem <= M, f"Ck_{j:02d}"
    # Objective
    model += sum_cost
    # Feasibility constraints, X[root][root] = 1 can be omitted else it ensures that X[root][root] must be 1
    for i in filter(lambda x: x != root, X):
        model += lp.lpSum(X[i].values()) == 1, f"Cf_{i:02d}"
    # Path-tree constraints
    if subchains:
        for pt in ((lp.lpSum(X[i][j] for i in tree.successors(v)) <= 1, f"Cp_{j:02d}_{v:02d}")
                   for v in X for j in X[v] if tree.succ[v]):
            model += pt
    # Latency constraint
    if L < math.inf:
        model += sum_lat <= L, LP_LAT
    else:
        # Add redundant constraint to implicitly calculate the latency value
        model += sum_lat >= 0, LP_LAT
    return model, X


def tree_mtx_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                          cp_end: int = None, delay: int = 1, subchains: bool = False,
                          solver: lp.LpSolver = None, timeout: int = None, **lpargs) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a tree based on the matrix ILP formulation.

    Block metrics are calculated based on serialized execution platform model.

    :param tree:        service graph annotated with node runtime(ms), memory(MB) and edge rate
    :param root:        root node of the graph
    :param M:           upper memory bound of the partition blocks (in MB)
    :param L:           latency limit defined on the critical path (in ms)
    :param cp_end:      tail node of the critical path in the form of subchain[root -> c_pend]
    :param delay:       invocation delay between blocks
    :param subchains:   only subchain blocks are considered (path-tree)
    :param solver:      specific solver class (default: COIN-OR CBC)
    :param timeout:     time limit in sec
    :param lpargs:      additional LP solver parameters
    :return:            tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    model, X = build_tree_mtx_model(tree, root, M, L, cpath, subchains, delay)
    status = model.solve(solver=solver if solver else lp.PULP_CBC_CMD(mip=True, msg=False, timeLimit=timeout, **lpargs))
    if status == lp.LpStatusOptimal:
        opt_cost, opt_lat = round(lp.value(model.objective), 0), round(lp.value(model.constraints[LP_LAT]), 0)
        return extract_subtrees_from_xmatrix(X), opt_cost, L + opt_lat if L < math.inf else opt_lat
    else:
        return INFEASIBLE


def recreate_subtrees_from_xmatrix(tree: nx.DiGraph, X: dict[int, dict[int, lp.LpVariable]]) -> T_PART:
    """
    Extract barrier nodes from variable matrix(dict-of-dict) and recreate partitioning blocks.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param X:       specific structure of decision variables
    :return:        calculated partitioning
    """
    return recreate_subtree_blocks(tree, barr=set(i for i in X if x_eval(X[i][i])))


def extract_subtrees_from_xmatrix(X: dict[int, dict[int, lp.LpVariable]]) -> T_PART:
    """
    Extract barrier nodes from variable matrix(dict-of-dict) and recreate partitioning blocks.
    
    :param X:   specific structure of decision variables
    :return:    calculated partitioning
    """
    return [[i for i in sorted(X) if j in X[i] and x_eval(X[i][j])] for j in sorted(X) if x_eval(X[j][j])]


########################################################################################################################


def all_tree_mtx_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                              cp_end: int = None, delay: int = 1, subchains: bool = False, solver: lp.LpSolver = None,
                              timeout: int = None, **lpargs) -> tuple[list[list[int]], int, int]:
    """
    Calculates all minimal-cost partitioning variations of a tree based on matrix ILP formulation.
    
    Block metrics are calculated based on serialized execution platform model.

    :param tree:        service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:        root node of the graph
    :param M:           upper memory bound of the partition blocks (in MB)
    :param L:           latency limit defined on the critical path (in ms)
    :param cp_end:      tail node of the critical path in the form of subchain[root -> c_pend]
    :param delay:       invocation delay between blocks
    :param subchains:   only subchain blocks are considered (path-tree)
    :param solver:      specific solver class (default: COIN-OR CBC)
    :param timeout:     time limit in sec
    :param lpargs:      additional LP solver parameters
    :return:            tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, set(ibacktrack_chain(tree, root, cpath)), M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    # Get model
    model, X = build_tree_mtx_model(tree, root, M, L, cpath, subchains, delay)
    # Init for min-cost results
    results, min_cost = [], math.inf
    solver = solver if solver else lp.PULP_CBC_CMD(mip=True, msg=False, timeLimit=timeout, **lpargs)
    # Iterate over the optimal solutions
    while model.solve(solver=solver) == lp.LpStatusOptimal:
        opt_cost, opt_lat = round(model.objective.value(), 0), round(model.constraints[LP_LAT].value(), 0)
        # If solution is not min-cost then exit else store the min-cost solution
        if min_cost < opt_cost:
            break
        else:
            min_cost = opt_cost
        results.append((extract_subtrees_from_xmatrix(X), opt_cost, L + opt_lat if L < math.inf else opt_lat))
        # Collect barrier nodes of the current optimal solution
        barr = [j for j in X if x_eval(X[j][j])]
        # Add extra constraint for excluding the current set of barrier nodes
        model += lp.lpSum(X[j][j] for j in barr) <= len(barr) - 1, f"Ca_{'_'.join(map(str, barr))}"
    # Return min-cost solutions
    return results if results else [INFEASIBLE]
